import json
import secrets
import time
from typing import Optional, Dict

import click

from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.utils import AwsClientWrapper


class CloudResourceService:
    _MCD_ROLE_PREFIX = 'monte-carlo-integration-role'
    _MCD_POLICY_PREFIX = 'monte-carlo-integration-cli-policy'
    _MCD_TAGS = [{'Key': 'MonteCarloData', 'Value': ''}]

    def __init__(self, config: Config, user_service: Optional[UserService] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None, aws_profile_override: Optional[str] = None):

        self._abort_on_error = True

        self._user_service = user_service or UserService(config=config)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=aws_profile_override or config.aws_profile,
                                                            region_name=config.aws_region)

    @manage_errors
    def create_role(self, path_to_policy_doc: str) -> None:
        """
        Creates a DC compatible role from the provided policy doc
        """
        current_time = str(time.time())
        external_id = self._generate_random_token()
        role_name = f'{self._MCD_ROLE_PREFIX}-{current_time}'
        policy_name = f'{self._MCD_POLICY_PREFIX}-{current_time}'

        try:
            policy = json.dumps(self._read_json(path_to_policy_doc))

            # use the AWS account id of collector, which may not necessarily be account id running the CLI
            account_id = self._user_service.active_collector['customerAwsAccountId']
        except json.decoder.JSONDecodeError as err:
            complain_and_abort(f'The provided policy is not valid JSON - {err}')
        except KeyError as err:
            complain_and_abort(f'Missing expected property ({err}). The collector may not have been deployed before')
        else:
            trust_policy = self._generate_trust_policy(account_id=account_id, external_id=external_id)
            role_arn = self._aws_wrapper.create_role(role_name=role_name, trust_policy=trust_policy, tags=self._MCD_TAGS)
            click.echo(f"Created role with ARN - '{role_arn}' and external id - '{external_id}'.")

            self._aws_wrapper.attach_inline_policy(role_name=role_name, policy_name=policy_name, policy_doc=policy)
            click.echo(f'Success! Attached provided policy.')

    @staticmethod
    def _generate_trust_policy(account_id: str, external_id: str) -> str:
        """
        Generates a DC compatible trust policy
        """
        return json.dumps(
            {
                'Version': '2012-10-17',
                'Statement': [
                    {
                        'Effect': 'Allow',
                        'Principal': {
                            'AWS': f'arn:aws:iam::{account_id}:root'
                        },
                        'Action': 'sts:AssumeRole',
                        'Condition': {
                            'StringEquals': {
                                'sts:ExternalId': external_id
                            }
                        }
                    }
                ]
            }
        )

    @staticmethod
    def _generate_random_token(length: Optional[int] = 16) -> str:
        """
        Generates a random token (e.g. for the external id)
        """
        return secrets.token_urlsafe(length)

    @staticmethod
    def _read_json(path: str) -> Dict:
        """
        Reads a JSON file from the path.
        """
        with open(path) as file:
            return json.load(file)  # loads for the purpose of validating



# Queries related to managing the collector

GENERATE_COLLECTOR_TEMPLATE = """
mutation generateCollectorTemplate {
  generateCollectorTemplate {
    dc {
      templateLaunchUrl
      stackArn
      customerAwsAccountId
      active
      apiGatewayId
    }
  }
}
"""



# Query responses
EXPECTED_HIVE_MYSQL_GQL_RESPONSE_FIELD = 'testDatabaseCredentials'
EXPECTED_HIVE_S3_GQL_RESPONSE_FIELD = 'testS3Credentials'
EXPECTED_HIVE_SQL_GQL_RESPONSE_FIELD = 'testHiveCredentials'
EXPECTED_PRESTO_SQL_GQL_RESPONSE_FIELD = 'testPrestoCredentials'
EXPECTED_PRESTO_S3_GQL_RESPONSE_FIELD = 'testS3Credentials'
EXPECTED_GLUE_GQL_RESPONSE_FIELD = 'testGlueCredentials'
EXPECTED_ATHENA_GQL_RESPONSE_FIELD = 'testAthenaCredentials'
EXPECTED_SPARK_GQL_RESPONSE_FIELD = 'testSparkCredentials'
EXPECTED_TOGGLE_EVENTS_GQL_RESPONSE_FIELD = 'toggleEventConfig'

# Available connections types
HIVE_MYSQL_CONNECTION_TYPE = 'hive-mysql'
HIVE_S3_CONNECTION_TYPE = 'hive-s3'
HIVE_SQL_CONNECTION_TYPE = 'hive'
PRESTO_SQL_CONNECTION_TYPE = 'presto'
PRESTO_S3_CONNECTION_TYPE = 'presto-s3'
GLUE_CONNECTION_TYPE = 'glue'
ATHENA_CONNECTION_TYPE = 'athena'
SPARK_CONNECTION_TYPE = 'spark'

# Available warehouse types
DATA_LAKE_WAREHOUSE_TYPE = 'data-lake'

# Job types
QL_JOB_TYPE = ['query_logs']

# Job limits
PRESTO_CATALOG_KEY = 'catalog_name'

# Certificate details
S3_CERT_MECHANISM = 'dc-s3'
PRESTO_CERT_PREFIX = 'certificates/presto/'
AWS_RDS_CA_CERT = 'https://s3.amazonaws.com/rds-downloads/rds-combined-ca-bundle.pem'

# Connections to friendly name (i.e. human presentable) map
GQL_TO_FRIENDLY_CONNECTION_MAP = {
    HIVE_MYSQL_CONNECTION_TYPE: 'Hive (metastore)',
    HIVE_S3_CONNECTION_TYPE: 'Hive (EMR logs)',
    HIVE_SQL_CONNECTION_TYPE: 'Hive (SQL)',
    PRESTO_SQL_CONNECTION_TYPE: 'Presto',
    PRESTO_S3_CONNECTION_TYPE: 'Presto (logs)',
    GLUE_CONNECTION_TYPE: 'Glue',
    ATHENA_CONNECTION_TYPE: 'Athena',
    SPARK_CONNECTION_TYPE: 'Spark (SQL)',
}



from dataclasses import dataclass
from typing import Optional

from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort
from montecarlodata.queries.user import GET_USER_QUERY
from montecarlodata.utils import GqlWrapper


@dataclass
class User:
    first_name: Optional[str] = None
    last_name: Optional[str] = None


class UserService:
    def __init__(self, config: Config, request_wrapper: Optional[GqlWrapper] = None):
        self._config = config
        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)

        self._user = self._request_wrapper.make_request(GET_USER_QUERY)  # get user info
        self._set_user_props()

    @property
    def user(self) -> User:
        """
        Returns basic user properties
        """
        return User(first_name=self._user['getUser'].get('firstName'), last_name=self._user['getUser'].get('lastName'))

    def _set_user_props(self) -> None:
        """
        Set relevant user props like DC and warehouse to instance
        """
        self.collectors = self._user['getUser']['account'].get('dataCollectors', [{}])
        self.active_collector = self.collectors[self._get_active_collector()]
        self.warehouses = self.active_collector.get('warehouses')

    def _get_active_collector(self) -> Optional[int]:
        """
        Get active collector - currently only one active collector per account is supported. Abort if None are found
        """
        for idx, collector in enumerate(self.collectors):
            if collector.get('active'):
                return idx
        complain_and_abort('No active collector found')



import json
import re
from typing import Dict, Optional, List

import boto3
import requests
from retry import retry

import montecarlodata.settings as settings
from montecarlodata.errors import manage_errors, complain_and_abort


class Wrapper:
    def __init__(self, abort_on_error: Optional[bool] = True):
        self._abort_on_error = abort_on_error


class GqlWrapper(Wrapper):
    def __init__(self, mcd_id: str, mcd_token: str):
        self._mcd_id = mcd_id
        self._mcd_token = mcd_token
        self._endpoint = settings.MCD_API_ENDPOINT

        super().__init__()

    def make_request(self, query: str, variables: Optional[Dict] = None) -> Optional[Dict]:
        """
        Make a GraphQl request to the MCD API. Aborts on error if set.
        """
        response = self._make_request(query=query, variables=variables)
        if response:
            if self._abort_on_error and response.get('errors'):
                complain_and_abort(response['errors'])
            return response['data']

    @staticmethod
    def convert_snakes_to_camels(dict_: Dict) -> Dict:
        """
        Converts dictionary keys from snake_case to camelCase as Gql is very opinionated
        """
        return {re.sub(r'_([a-z])', lambda x: x.group(1).upper(), k): v for k, v in dict_.items()}

    @manage_errors
    def _make_request(self, query: str, variables: Optional[Dict] = None) -> Optional[Dict]:
        headers = {'x-mcd-id': self._mcd_id, 'x-mcd-token': self._mcd_token, 'Content-Type': 'application/json'}
        payload = {'query': query, 'variables': variables or {}}

        response = self._post(headers=headers, payload=payload)
        response.raise_for_status()

        return json.loads(response.text)

    @retry(tries=3, delay=0.2, backoff=2, max_delay=1)
    def _post(self, headers: Dict, payload: Dict) -> requests.request:
        return requests.post(self._endpoint, json=payload, headers=headers)


class AwsClientWrapper(Wrapper):
    _CF_UPDATE_WAITER = 'stack_update_complete'
    _DEFAULT_WAITER_CONFIG = {'Delay': 5, 'MaxAttempts': 720}
    _STACK_UPDATE_STATE = 'UPDATE_COMPLETE'

    def __init__(self, profile_name: Optional[str] = None, region_name: Optional[str] = None,
                 session: Optional[boto3.session.Session] = None):
        super().__init__()

        self._session = session or self._get_session(profile_name=profile_name, region_name=region_name)
        self._test_if_region_is_set()

    @manage_errors
    def get_stack_details(self, stack_id: str) -> Dict:
        """
        Retrieve stack details (description) from DC
        """
        return self._session.client('cloudformation').describe_stacks(StackName=stack_id)

    @manage_errors
    def get_stack_outputs(self, stack_id: str) -> Dict:
        """
        Convenience utility to retrieve stack outputs from DC
        """
        return self.get_stack_details(stack_id)['Stacks'][0]['Outputs']

    @manage_errors
    def get_stack_parameters(self, stack_id: str) -> List[Dict]:
        """
        Convenience utility to retrieve stack parameters from the DC
        """
        return self.get_stack_details(stack_id)['Stacks'][0].get('Parameters')

    @manage_errors
    def upgrade_stack(self, stack_id: str, template_link: str, parameters: List[Dict],
                      capabilities: Optional[str] = 'CAPABILITY_IAM') -> bool:
        """
        Attempts to upgrade the DC stack - returning True on success
        """
        self._session.client('cloudformation').update_stack(StackName=stack_id, TemplateURL=template_link,
                                                            Capabilities=[capabilities], Parameters=parameters)
        self._session.client('cloudformation').get_waiter(self._CF_UPDATE_WAITER).wait(StackName=stack_id,
                                                                                       WaiterConfig=self._DEFAULT_WAITER_CONFIG)

        if self.get_stack_details(stack_id)['Stacks'][0]['StackStatus'] == self._STACK_UPDATE_STATE:
            return True
        return False

    @manage_errors
    def deploy_gateway(self, gateway_id: str, stage: Optional[str] = 'prod') -> None:
        """
        Attempts to deploy the DC gateway
        """
        self._session.client('apigateway').create_deployment(restApiId=gateway_id, stageName=stage)

    @manage_errors
    def upload_file(self, bucket_name: str, object_name: str, file_path: str) -> None:
        """
        Upload a file to s3://bucket/object from file path
        """
        self._session.client('s3').upload_file(file_path, bucket_name, object_name)

    @manage_errors
    def create_role(self, role_name: str, trust_policy: str, tags: Optional[List] = None) -> str:
        """
        Creates an IAM role and returns the ARN on success
        """
        return self._session.client('iam').create_role(RoleName=role_name,
                                                       AssumeRolePolicyDocument=trust_policy, Tags=tags)['Role']['Arn']

    @manage_errors
    def attach_inline_policy(self, role_name: str, policy_name: str, policy_doc: str) -> str:
        """
        Attaches an inline policy to an existing role
        """
        return self._session.client('iam').put_role_policy(RoleName=role_name, PolicyName=policy_name,
                                                           PolicyDocument=policy_doc)

    @manage_errors
    def _test_if_region_is_set(self) -> None:
        """
        Test if region is set by creating a client. Abort on failure
        """
        self._session.client('cloudformation')  # throws NoRegion if not set

    @manage_errors
    def _get_session(self, profile_name: str, region_name: str) -> Optional[boto3.session.Session]:
        """
        Create a session using named profile/region if set. Uses AWS defaults if not set.
        """
        session_args = (('profile_name', profile_name), ('region_name', region_name))
        return boto3.session.Session(**{k: v for k, v in session_args if v})  # throws ProfileNotFound if invalid



import json
import os
from typing import Optional, Dict, List

import click

from montecarlodata.common.common import normalize_gql
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort
from montecarlodata.integrations.onboarding.fields import S3_CERT_MECHANISM, GQL_TO_FRIENDLY_CONNECTION_MAP, \
    DATA_LAKE_WAREHOUSE_TYPE
from montecarlodata.queries.onboarding import ADD_CONNECTION_MUTATION
from montecarlodata.utils import GqlWrapper, AwsClientWrapper


class BaseOnboardingService:

    def __init__(self, config: Config, user_service: Optional[UserService] = None,
                 request_wrapper: Optional[GqlWrapper] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None):

        self._abort_on_error = True  # Aborts methods with deco on unhandled error
        self._dc_outputs = None

        self._user_service = user_service or UserService(config=config)
        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=config.aws_profile,
                                                            region_name=config.aws_region)

    def onboard(self, validation_query: str, validation_response: str, connection_type: str,
                job_types: Optional[List[str]] = None, job_limits: Optional[Dict] = None, **kwargs):
        """
        Convenience wrapper to validate and add a connection.
        """
        temp_path = self._validate_connection(query=validation_query, response_field=validation_response, **kwargs)
        self._add_connection(temp_path=temp_path, connection_type=connection_type, job_types=job_types,
                             job_limits=job_limits)

    def handle_cert(self, cert_prefix: str, options: Dict) -> None:
        """
        Handles cert payload from either an s3 path or file. Uploading the latter
        Options is updated if successful.
        """
        if options.get('cert_file') is not None:
            bucket_name = self._get_dc_property(prop='PrivateS3BucketArn').split(':')[5]  # get name from arn
            object_name = os.path.join(cert_prefix, os.path.basename(options['cert_file']))
            self._aws_wrapper.upload_file(bucket_name=bucket_name, object_name=object_name,
                                          file_path=options['cert_file'])

            click.echo(f"Uploaded '{options['cert_file']}' to s3://{bucket_name}/{object_name}")
            options['cert_s3'] = object_name

        if options.get('cert_s3') is not None:
            # reformat to generic options and specify a mechanism
            options['ssl_options'] = {'mechanism': S3_CERT_MECHANISM, 'cert': options.pop('cert_s3')}
        options.pop('cert_file', None)

    def _validate_connection(self, query: str, response_field: str, **kwargs) -> str:
        """
        Validate connection before adding using the expected gql response_field (e.g. testPrestoCredentials)
        """
        variables = self._request_wrapper.convert_snakes_to_camels(kwargs)
        temp_path = self._request_wrapper.make_request(
            query=query, variables=variables).get(response_field, {}).get('key')

        if temp_path is not None:
            click.echo('Connection validated!')
            return temp_path
        complain_and_abort('Connection failed!')

    def _add_connection(self,
                        temp_path: str,
                        connection_type: str,
                        job_types: Optional[List[str]] = None,
                        warehouse_type: Optional[str] = DATA_LAKE_WAREHOUSE_TYPE,
                        job_limits: Optional[Dict] = None) -> bool:
        """
        Add connection and setup any associated jobs
        """
        connection_request = {'key': temp_path, 'connectionType': connection_type}
        num_of_warehouses = len(self._user_service.warehouses)

        # If no warehouse has been created for this account specify the type, otherwise pass the existing warehouse id
        if num_of_warehouses == 0:
            connection_request['createWarehouseType'] = warehouse_type
        elif num_of_warehouses == 1 and normalize_gql(self._user_service.warehouses[0].get('connectionType')) == warehouse_type:
            connection_request['dwId'] = self._user_service.warehouses[0]['uuid']
        else:
            complain_and_abort('More than one warehouse is not supported')

        if job_types:
            # specify job types if set
            connection_request['jobTypes'] = job_types
        if job_limits:
            connection_request['jobLimits'] = json.dumps(job_limits)

        response = self._request_wrapper.make_request(query=ADD_CONNECTION_MUTATION, variables=connection_request)
        connection_id = response.get('addConnection', {}).get('connection', {}).get('uuid')
        if connection_id is not None:
            click.echo(f"Success! Added connection for "
                       f"{GQL_TO_FRIENDLY_CONNECTION_MAP.get(connection_type, connection_type.capitalize())}.")
            return True
        complain_and_abort('Failed to add connection!')

    def _get_dc_property(self, prop: str) -> Optional[str]:
        """
        Retrieve property from DC stack outputs
        """
        self._dc_outputs = self._dc_outputs or self._aws_wrapper.get_stack_outputs(
            self._user_service.active_collector['stackArn'])  # cache lookup
        for output in self._dc_outputs:
            if output['OutputKey'] == prop:
                return output['OutputValue']



import click
import click_config_file

import montecarlodata.settings as settings
from montecarlodata.collector.management import CollectorManagementService
from montecarlodata.tools import validate_json_callback


@click.group(help='Manage the data collector.')
def collector():
    """
    Group for any collector related subcommands
    """
    pass


@collector.command(help='Get link to the latest template. For initial deployment or manually upgrading.')
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def get_template(ctx):
    """
    Get the latest template for this account
    """
    CollectorManagementService(config=ctx['config']).echo_template()


@collector.command(help='Upgrade to the latest version.')
@click.pass_obj
@click.option('--params', required=False, default=None, callback=validate_json_callback,
              help="""
              Parameters key,value pairs as JSON. If a key is not specified the existing (or default) value is used.
              \b
              \n
              E.g. --params '{"CreateEventInfra":"True"}'
              """)  # \b disables wrapping
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def upgrade(ctx, params):
    """
    Upgrade the collector for this account
    """
    CollectorManagementService(config=ctx['config']).upgrade_template(new_params=params)



"""
Command line tools
"""
import json

import click


def add_common_options(options):
    """
    Convenience decorator for shared options (i.e. options that are common across commands)
    """

    def _add_common_options(function):
        for option in options[::-1]:
            function = option(function)
        return function

    return _add_common_options


def validate_json_callback(ctx_, param_, value):
    """
    Convenience callback to help validate (and load) JSON in option strings
    """
    try:
        return json.loads(value)
    except json.decoder.JSONDecodeError as err:
        raise click.BadParameter(f'Malformed JSON - {err}')
    except TypeError:
        return value



# Queries related to onboarding

TEST_PRESTO_CRED_MUTATION = """
mutation testPrestoCredentials($catalog:String, $host:String, $httpScheme:String, $password:String, $port:Int, $schema:String, $sslOptions:SslInputOptions, $user:String) {
  testPrestoCredentials(catalog:$catalog, host:$host, httpScheme:$httpScheme, password:$password, port:$port, schema:$schema, sslOptions: $sslOptions, user:$user) {
    key
  }
}
"""

TEST_S3_CRED_MUTATION = """
mutation tests3Credentials($bucket:String, $prefix:String, $assumableRole:String, $externalId:String, $connectionType:String) {
  testS3Credentials(bucket:$bucket, prefix:$prefix, assumableRole:$assumableRole, externalId:$externalId, connectionType:$connectionType) {
    key
  }
}
"""

TEST_DATABASE_CRED_MUTATION = """
mutation testDatabaseCredentials($assumableRole:String, $connectionType:String, $dbName:String, $externalId:String, $host:String, $password:String, $port:Int, $sslOptions:SslInputOptions, $user:String) {
  testDatabaseCredentials(assumableRole:$assumableRole, connectionType:$connectionType, dbName:$dbName, externalId:$externalId, host:$host, password:$password, port:$port, sslOptions:$sslOptions, user:$user) {
    key
  }
}
"""

TEST_HIVE_SQL_CRED_MUTATION = """
mutation testHiveCredentials($database:String, $host:String, $port:Int, $username:String) {
  testHiveCredentials(database:$database, host:$host, port:$port, username:$username) {
    key
    success
  }
}
"""

TEST_GLUE_CRED_MUTATION = """
mutation testGlueCredentials($assumableRole:String, $externalId:String, $awsRegion:String) {
  testGlueCredentials(assumableRole:$assumableRole, externalId:$externalId, awsRegion:$awsRegion) {
    key
  }
}
"""

TEST_ATHENA_CRED_MUTATION = """
mutation testAthenaCredentials($assumableRole:String, $catalog:String, $externalId:String, $workgroup:String, $awsRegion:String) {
  testAthenaCredentials(assumableRole:$assumableRole, catalog:$catalog, externalId:$externalId, workgroup:$workgroup, awsRegion:$awsRegion) {
    key
  }
}
"""

TEST_SPARK_BINARY_MODE_CRED_MUTATION = """
mutation testSparkCredentials($database:String!, $host:String!, $port:Int!, $username:String!, $password:String!) {
  testSparkCredentials(binaryMode: {database:$database, host:$host, port:$port, username:$username, password:$password}) {
    key
  }
}
"""

TEST_SPARK_HTTP_MODE_CRED_MUTATION = """
mutation testSparkCredentials($url:String!, $username:String!, $password:String!) {
  testSparkCredentials(httpMode: {url:$url, username:$username, password:$password}) {
    key
  }
}
"""

TEST_SPARK_DATABRICKS_CRED_MUTATION = """
mutation testSparkCredentials($workspaceUrl:String!, $workspaceId:String!, $clusterId:String!, $token:String!) {
  testSparkCredentials(databricks: {workspaceUrl:$workspaceUrl, workspaceId:$workspaceId, clusterId:$clusterId, token:$token}) {
    key
  }
}
"""

TOGGLE_METADATA_EVENT_MUTATION = """
mutation toggleEventMetadataConfig($dwId:UUID!, $enable:Boolean!, $connectionType:String!) {
  toggleEventConfig(dwId:$dwId, enable:$enable, connectionType:$connectionType, eventType: s3_metadata_events){
    success
  }
}
"""

ADD_CONNECTION_MUTATION = """
mutation addConnection($connectionType:String!, $createWarehouseType:String, $dwId:UUID, $jobTypes:[String], $key:String!, $jobLimits:JSONString) {
  addConnection(connectionType:$connectionType, createWarehouseType:$createWarehouseType, dwId:$dwId, jobTypes:$jobTypes, key:$key, jobLimits:$jobLimits){
    connection {
      uuid
    }
  }
}
"""



import copy
from typing import Optional, Dict, List
from urllib import parse

import click

from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, complain_and_abort
from montecarlodata.queries.collector import GENERATE_COLLECTOR_TEMPLATE
from montecarlodata.utils import GqlWrapper, AwsClientWrapper


class CollectorManagementService:
    def __init__(self, config: Config, user_service: Optional[UserService] = None,
                 request_wrapper: Optional[GqlWrapper] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None):
        self._abort_on_error = True

        self._user_service = user_service or UserService(config=config)
        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id, mcd_token=config.mcd_token)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=config.aws_profile,
                                                            region_name=config.aws_region)

    @manage_errors
    def echo_template(self) -> None:
        """
        Echos the most recent template for this account
        """
        click.echo(self._get_template_url_from_launch(self._generate_collector_template().get('templateLaunchUrl')))

    @manage_errors
    def upgrade_template(self, new_params: Optional[Dict] = None) -> None:
        """
        Upgrades the DC attached to this account
        """
        collector_details = self._generate_collector_template()
        try:
            stack_arn = collector_details['stackArn']
            is_active = collector_details['active']
            gateway_id = collector_details['apiGatewayId']
            template_url = self._get_template_url_from_launch(collector_details['templateLaunchUrl'])

            if not is_active:
                complain_and_abort('Cannot upgrade an inactive collector. Please contact Monte Carlo')
        except KeyError as err:
            complain_and_abort(f'Missing expected property ({err}). The collector may not have been deployed before')
        else:
            click.echo(f'Updating \'{stack_arn}\'')
            self._upgrade(stack_arn=stack_arn, template_url=template_url, gateway_id=gateway_id, new_params=new_params)
            click.echo('Upgrade complete! Have a nice day!')

    def _upgrade(self, stack_arn: str, template_url: str, gateway_id: str, new_params: Dict):
        """
        Upgrade the collector stack and deploy the gateway
        """
        parameters = self._build_param_list(self._aws_wrapper.get_stack_parameters(stack_id=stack_arn), new_params)
        if self._aws_wrapper.upgrade_stack(stack_id=stack_arn, template_link=template_url, parameters=parameters):
            self._aws_wrapper.deploy_gateway(gateway_id=gateway_id)
        else:
            complain_and_abort('Failed to upgrade. Please review CF events for details')

    def _generate_collector_template(self) -> Optional[Dict]:
        """
        Generate the latest template and returns any associated dc properties (i.e. from the DataCollectorModel)
        """
        response = self._request_wrapper.make_request(GENERATE_COLLECTOR_TEMPLATE)
        return response['generateCollectorTemplate'].get('dc')

    @staticmethod
    def _get_template_url_from_launch(launch_url: str) -> str:
        """
        Extract the template url from expected launch url structure
        """
        return parse.parse_qs(parse.urlparse(launch_url).fragment).get('templateURL', [])[0]

    @staticmethod
    def _build_param_list(existing_params: List[Dict], new_params: Optional[Dict] = None) -> List[Dict]:
        """
        Get a list of parameters by replacing (or extending) any new params into the existing stack params
        """
        existing_params = copy.deepcopy(existing_params)
        new_params = copy.deepcopy(new_params)

        for param in existing_params or []:
            if new_params and new_params.get(param['ParameterKey']) is not None:
                param['ParameterValue'] = new_params[param['ParameterKey']]
                param['UsePreviousValue'] = False
                del new_params[param['ParameterKey']]
            else:
                del param['ParameterValue']
                param['UsePreviousValue'] = True
        if new_params:
            # Handle any completely new params (i.e. those that were not in the existing struct)
            # Can be found for params added between versions
            existing_params.extend([{'ParameterKey': k, 'ParameterValue': v} for k, v in new_params.items()])

        return existing_params
